using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

using System.Reflection;
using System.IO;
using System.Threading;
using System.Diagnostics;
using Microsoft.Win32;

namespace AG3Ghost
{
    public partial class AG3Ghost : Form
    {
        static Copier copier;
        const string userRoot = "HKEY_CURRENT_USER";
        const string lmRoot = "HKEY_LOCAL_MACHINE";
        const string subkey = @"\Software\illusion\JS3";
        const string subkey64 = @"\Software\Wow6432Node\illusion\JS3";
        const string keyName = userRoot + subkey;
        const string keyName64 = userRoot + subkey64;
        const string keyNameLm = lmRoot + subkey;
        const string keyNameLm64 = lmRoot + subkey64;
        const string workspaceDir = @"\AG3GhostWorkspace\";
        const string workspaceXxDir = @"\js3_00_02_00\";
        string ag3dir;
        string workspacePath;
        string ppFile;
        static bool workSpaceReady = true;
        static bool workSpaceFailed = false;
        ManualResetEvent manualEvent;
        //ManualResetEvent workspaceEvent;

        public AG3Ghost()
        {
            InitializeComponent();
            
            this.Load += new EventHandler(AG3Ghost_Load);
        }

        #region Prerequisite checking
        bool IsAG3ProcessRunning()
        {
            bool process = false;
            if (IsProcessOpen("AG3_Make") || IsProcessOpen("JS3_Make")
                || IsProcessOpen("AG3_Play")
                || IsProcessOpen("JS3_Play"))
            {
                MessageBox.Show("AG3 Ghost has determined that you have the game open.\n\nClose both the game and maker before running AG3 Ghost.", "AG3 Game or Make Running", MessageBoxButtons.OK, MessageBoxIcon.Error);                
                process = true;
                this.Close();                
            }
            return process;
        }

        bool IsAG3Installed()
        {
            bool installed = true;
            ag3dir = (string)Registry.GetValue(keyName, "INSTALLDIR", null);
            if (String.IsNullOrEmpty(ag3dir))
                ag3dir = (string)Registry.GetValue(keyName64, "INSTALLDIR", null);
            if (String.IsNullOrEmpty(ag3dir))
                ag3dir = (string)Registry.GetValue(keyNameLm, "INSTALLDIR", null);
            if (String.IsNullOrEmpty(ag3dir))
                ag3dir = (string)Registry.GetValue(keyNameLm64, "INSTALLDIR", null);

            if (String.IsNullOrEmpty(ag3dir))
            {
                installed = false;
                MessageBox.Show("AG3 Ghost was not able to find your AG3 game. There was no entry in the Windows Registry.\n\nMake sure that the game is properly installed before running AG3 Ghost.", "AG3 Game not found", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }            
            else if ((!File.Exists(ag3dir + "\\AG3_Make.exe") && !File.Exists(ag3dir + "\\JS3_Make.exe")))
            {
                installed = false;
                MessageBox.Show("AG3 Ghost was not able to find either AG3_Make.exe or JS3_Make.exe in your AG3 game folder.\n\nAG3 Ghost requires one of these to be present to be able to run.\n\nRename your favorite maker to 'AG3_Make.exe' or 'JS3_Make.exe' to run AG3 Ghost.", "AG3_Make.exe or JS3_Make.exe not found", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

            if (!installed)
                this.Close();

            return installed;                
        }

        bool IsEnoughFreeSpace()
        {
            bool enough = true;
            try
            {
                FileInfo fi = new FileInfo(ag3dir + "\\data\\js3_00_02_00.pp");
                DriveInfo di = new DriveInfo(Directory.GetDirectoryRoot(ag3dir));
                //long len = fi.Length * 3;
                //long len2 = fi.Length * 2 + 100000000;
                if ((fi.Length * 2 + 100000000) > di.AvailableFreeSpace)
                {
                    enough = false;
                    MessageBox.Show("AG3 Ghost has determined that you don't have enough free disk space to run AG3 Ghost.\n\nMake sure that you have at least two times the size of your js3_00_02_00.pp free on the drive where you installed AG3.", "Not enough available space", MessageBoxButtons.OK, MessageBoxIcon.Error);                    
                }
            }
            catch (Exception e)
            {
                MessageBox.Show("AG3 Ghost encountered a problem trying to determine if you have enough free disk space for the workspace.\n\nAG3 Ghost is unable to continue. (press ctrl+C to copy this message)\nException: " + e.Message, "Problem determining free space", MessageBoxButtons.OK, MessageBoxIcon.Error);
                enough = false;
            }

            if (!enough)
                this.Close();

            return enough;
        }

        public bool IsProcessOpen(string name)
        {
            foreach (Process clsProcess in Process.GetProcesses())
            {
                if (clsProcess.ProcessName.Contains(name))
                {
                    return true;
                }
            }
            return false;
        } 

        #endregion Prerequisite checking

        void AG3Ghost_Load(object sender, EventArgs e)
        {
            if (IsAG3ProcessRunning())
                return;
                        
            if (!IsAG3Installed())
                return;

            if (!IsEnoughFreeSpace())
                return;

            this.FormClosing += new FormClosingEventHandler(AG3Ghost_FormClosing);
            copier = new Copier(this);
            copier.ProgressChanged += new EventHandler(copier_ProgressChanged);

            
            workspacePath = ag3dir + @"\data\" + @"\AG3GhostWorkspace\";
            //workspaceEvent = new ManualResetEvent(false);
            Application.DoEvents();
            SetupWorkSpace(workspacePath);
        }
        
        void AG3Ghost_FormClosing(object sender, FormClosingEventArgs e)
        {
            Restore();
            DialogResult result = MessageBox.Show("Do you wish to delete the AG3 Ghost workspace? \n\nLeaving the workspace allows AG3 Ghost to start faster.\nClick Yes if you do not plan to run AG3 Ghost again sometime soon."
                , "Delete AG3 Ghost workspace", MessageBoxButtons.YesNo, MessageBoxIcon.Question);
            if (result == DialogResult.Yes)
                DeleteWorkspace();
        }
        
        void DeleteWorkspace()
        {
            try
            {
                Directory.SetCurrentDirectory(ag3dir);
                if (Directory.Exists(workspacePath))
                    Directory.Delete(workspacePath, true);
            }
            catch
            {
            }
        }

        void Restore()
        {
            if (File.Exists(ag3dir + "\\data\\js3_00_02_00.pp.AG3Ghost"))
            {
                File.Delete(ag3dir + "\\data\\js3_00_02_00.pp");
                File.Move(ag3dir + "\\data\\js3_00_02_00.pp.AG3Ghost", ag3dir + "\\data\\js3_00_02_00.pp");
            }
        }

        void SetupWorkSpace(string ws)
        {            
            workSpaceReady = false;
            Directory.CreateDirectory(ws);
            InstallPPExtractor(workspacePath);
            progressBar1.Value = 50;
            Thread thread = new Thread(AG3Ghost.CreateWorkSpace);
            thread.Start(ag3dir + @"\data\");
        }
        
        void copier_ProgressChanged(object sender, EventArgs e)
        {
            InvokeStatusChanged("Creating AG3 Ghost workspace...");
            InvokeSetProgress(50 + (int)copier.Progress);
            //InvokeAddToProgress((int)copier.Progress);
            if (copier.Progress == 100)
            {
                labelStatus.Text = "AG3 Ghost workspace ready.";
                workSpaceReady = true;
                //workspaceEvent.Set();
            }
        }

        private void buttonBrowse_Click(object sender, EventArgs e)
        {            
            OpenFileDialog openFileDialog = new OpenFileDialog();
            openFileDialog.RestoreDirectory = true;
            openFileDialog.ValidateNames = true;
            openFileDialog.InitialDirectory = ag3dir + @"\data\save\m_cha";
            openFileDialog.Filter = "pp files (*.pp)|*.pp|All files (*.*)|*.*";
            openFileDialog.FilterIndex = 0;            

            openFileDialog.ShowDialog();
                        
            ppFile = openFileDialog.FileName;
            if (String.IsNullOrEmpty(ppFile)) return;

            //workspaceEvent.WaitOne();

            //System.Windows.Forms.Timer timer = new Timer(
            while (!workSpaceReady)
            {
                Application.DoEvents();
                labelStatus.Text = "Waiting for the workspace to be ready.";
                if (workSpaceFailed)
                {
                    labelStatus.Text = "Failed creating the workspace.";
                    return;
                }
            }

            this.buttonBrowse.Enabled = false;

            manualEvent = new ManualResetEvent(false);

            labelStatus.Text = "Extracting character...";
            progressBar1.Value = 25;
            bool canContinue = GetCharXxFiles(ppFile);
            //labelStatus.Text = "Ready.";
            if (!canContinue)
            {
                labelStatus.Text = "";
                progressBar1.Value = 0;
                return;
            }
            //remember to test if it is possible to load another pp file while working on one!!!
            labelStatus.Text = "Modifying game files...";
            progressBar1.Value = 35;
            manualEvent.WaitOne();
            MergePP();
        }
                
        void ReadyToRun()
        {
            this.buttonBrowse.Enabled = true;
            InvokeSetProgress(150);
            InvokeStatusChanged("All set. Launching AG3_Make...");
            this.Enabled = false;
            RunAg3Make();
        }

        bool GetCharXxFiles(string charPath)
        {
            bool succeded = true;
            if (!ExtractPP(charPath))
                return false;
            manualEvent.WaitOne();
            try
            {
                string dir = Path.GetDirectoryName(charPath);
                string xxPath = dir + "\\" + Path.GetFileNameWithoutExtension(charPath);
                if (!Directory.Exists(xxPath)) throw new Exception();
                File.Copy(xxPath + "\\head.xx", workspacePath + @"\js3_00_02_00\a01_01.xx", true);
                File.Copy(xxPath + "\\head.xx", workspacePath + @"\js3_00_02_00\d01_01.xx", true);
                File.Copy(xxPath + "\\body.xx", workspacePath + @"\js3_00_02_00\a15_01.xx", true);                
                File.Copy(xxPath + "\\body.xx", workspacePath + @"\js3_00_02_00\d15_01.xx", true);

                File.Copy(xxPath + "\\hairF.xx", workspacePath + @"\js3_00_02_00\a12_01_01.xx", true);
                File.Copy(xxPath + "\\hairB.xx", workspacePath + @"\js3_00_02_00\a12_02_01.xx", true);
                File.Copy(xxPath + "\\hairS.xx", workspacePath + @"\js3_00_02_00\a12_03_01.xx", true);
                Directory.Delete(xxPath, true);
            }
            catch
            {
                succeded = false;
                MessageBox.Show("AG3 Ghost failed to apply the character file.\n\nMake sure that you select a valid character pp file", "Character file problem", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            manualEvent.Set();

            return succeded;
        }
                
        static void CreateWorkSpace(object ag3dataDir)
        {
            try
            {                
                Directory.CreateDirectory((string)ag3dataDir + workspaceDir + workspaceXxDir);
                if (!File.Exists((string)ag3dataDir + workspaceDir + "js3_00_02_00.pp"))
                    copier.XCopy((string)ag3dataDir + "js3_00_02_00.pp", (string)ag3dataDir + workspaceDir + "js3_00_02_00.pp");
                else
                    copier.Progress = 100;
            }
            catch (Exception e)
            {
                AG3Ghost.workSpaceFailed = false;
                MessageBox.Show("Error creating workspace. " + e.Message);
            }            
        }

        void RunAg3Make()
        {
            Process ag3MakeProcess = new Process();
            if (File.Exists(ag3dir + "\\AG3_Make.exe"))
                ag3MakeProcess.StartInfo.FileName = ag3dir + "\\AG3_Make.exe";            
            else
                ag3MakeProcess.StartInfo.FileName = ag3dir + "\\JS3_Make.exe";
            Directory.SetCurrentDirectory(workspacePath);
            ag3MakeProcess.StartInfo.Arguments = "/nobackup /merge js3_00_02_00";

            ag3MakeProcess.EnableRaisingEvents = true;
            ag3MakeProcess.Exited += new EventHandler(ag3MakeProcess_Exited);

            labelStatus.Text = "Waiting for AG3_Make to close...";

            ag3MakeProcess.Start();
        }

        void ag3MakeProcess_Exited(object sender, EventArgs e)
        {
            this.Invoke(new ThisEnabledDelegate(ThisEnabled), new object[] { true });
            InvokeSetProgress(0);
            InvokeStatusChanged("Game files restored.");
            Restore();
        }

        delegate void ThisEnabledDelegate(bool enabled);

        void ThisEnabled(bool enabled)
        {
            this.Enabled = enabled;
        }

        bool ExtractPP(string ppFilePath)
        {
            bool succeeded = true;
            try
            {
                Process extractProcess = new Process();
                extractProcess.StartInfo.FileName = workspacePath + @"\\PPExtractor.exe";
                extractProcess.StartInfo.WindowStyle = ProcessWindowStyle.Hidden;
                string dir = Path.GetDirectoryName(ppFilePath);
                Directory.SetCurrentDirectory(dir);
                string ppFile = Path.GetFileName(ppFilePath);
                extractProcess.StartInfo.Arguments = ppFile;
                extractProcess.EnableRaisingEvents = true;
                extractProcess.Exited += new EventHandler(extractProcess_Exited);

                extractProcess.Start();
            }
            catch
            {
                succeeded = false;
                MessageBox.Show("There was a problem extracting the character file.\n\nMake sure that it is a valid AG3 character pp file", "Problem with character file", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            return succeeded;
        }

        void extractProcess_Exited(object sender, EventArgs e)
        {
            manualEvent.Set();
        }

        void MergePP()
        {
            Process mergeProcess = new Process();
            mergeProcess.StartInfo.FileName = workspacePath + @"\\PPExtractor.exe";
            mergeProcess.StartInfo.WindowStyle = ProcessWindowStyle.Hidden;
            Directory.SetCurrentDirectory(workspacePath);            
            mergeProcess.StartInfo.Arguments = "/nobackup /merge js3_00_02_00";

            mergeProcess.EnableRaisingEvents = true;
            mergeProcess.Exited += new EventHandler(mergeProcess_Exited);

            mergeProcess.Start();
        }

        void mergeProcess_Exited(object sender, EventArgs e)
        {
            InvokeAddToProgress(110);
            InstallGhostPP();
        }

        void InstallGhostPP()
        {
            if (!File.Exists(ag3dir + "\\data\\js3_00_02_00.pp.AG3Ghost"))
                File.Move(ag3dir + "\\data\\js3_00_02_00.pp", ag3dir + "\\data\\js3_00_02_00.pp.AG3Ghost");
            if (!File.Exists(ag3dir + "\\data\\js3_00_02_00.pp"))
                File.Copy(workspacePath + "js3_00_02_00.pp", ag3dir + "\\data\\js3_00_02_00.pp");
            this.Invoke(new ReadyToRunDelegate(ReadyToRun));
        }
        
        delegate void ReadyToRunDelegate();
                
        void InstallPPExtractor(string path)
        {
            if (File.Exists(path + "\\PPExtractor.exe")) return;

            BinaryWriter sw = null;
            Stream memStream = null;
            FileStream fs = null;
            try
            {
                fs = new FileStream(path + "\\PPExtractor.exe", FileMode.CreateNew);
                memStream = Assembly.GetExecutingAssembly().GetManifestResourceStream("AG3Ghost.res.PPExtractor.dat");//new MemoryStream(global::ZoomModSettings.Properties.Resources.camera_high);
                byte[] buffer = new byte[memStream.Length];
                memStream.Read(buffer, 0, buffer.Length);
                sw = new BinaryWriter(fs);

                sw.Write(buffer);
                sw.Close();
            }
            catch (Exception e)
            {
                workSpaceFailed = false;
                MessageBox.Show("There was an error when trying to copy PPExtractor.\n" + e.Message);
            }
            finally
            {
                if (sw != null)
                    sw.Close();
                if (fs != null)
                    fs.Close();
                if (memStream != null)
                    memStream.Close();

            }
        }
                
        #region Invoking

        public delegate void ProgressChangedDelegate(int progress);

        void AddToProgress(int progress)
        {
            progressBar1.Value += progress;
            if (progressBar1.Value == 150)
                return;
        }

        void InvokeAddToProgress(int progress)
        {
            object[] myArray = new object[1];

            myArray[0] = progress;

            if (this.InvokeRequired)
            {
                this.Invoke(new ProgressChangedDelegate(AddToProgress), myArray);
            }
            else
                AddToProgress(progress);
        }

        void SetProgress(int progress)
        {
            progressBar1.Value = progress;
        }

        void InvokeSetProgress(int progress)
        {
            object[] myArray = new object[1];

            myArray[0] = progress;

            if (this.InvokeRequired)
            {
                this.Invoke(new ProgressChangedDelegate(SetProgress), myArray);
            }
            else
                SetProgress(progress);
        }

        public delegate void StatusChangedDelegate(string progress);

        void StatusChanged(string status)
        {
            labelStatus.Text = status;
        }

        void InvokeStatusChanged(string status)
        {
            object[] myArray = new object[1];

            myArray[0] = status;

            if (this.InvokeRequired)
            {
                this.Invoke(new StatusChangedDelegate(StatusChanged), myArray);
            }
            else
                StatusChanged(status);
        }
        #endregion Invoking

        ToolTip toolTip;            
        ToolTip Tip
        {
            get 
            {
                if (toolTip == null)
                {
                    toolTip = new ToolTip();
                    //toolTip.IsBalloon = true;                    
                    //toolTip.ToolTipTitle = "Why is this useful?";                    
                }
                return toolTip;
            }
        }
                
        private void linkLabel1_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            Tip.Show(Properties.Resources.ToolTip, this, progressBar1.Right, pictureBox1.Top, 60000);
        }
    }
}