using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace AG3Ghost
{
    internal class Copier
    {
        Form theForm;
        internal Copier(Form theForm)
        {
            this.theForm = theForm;
        }

        [DllImport("kernel32.dll", SetLastError = true, CharSet = CharSet.Auto)]
        [return: MarshalAs(UnmanagedType.Bool)]
        static extern bool CopyFileEx(string lpExistingFileName, string lpNewFileName,
           CopyProgressRoutine lpProgressRoutine, IntPtr lpData, ref Int32 pbCancel,
           CopyFileFlags dwCopyFlags);

        delegate CopyProgressResult CopyProgressRoutine(
        long TotalFileSize,
        long TotalBytesTransferred,
        long StreamSize,
        long StreamBytesTransferred,
        uint dwStreamNumber,
        CopyProgressCallbackReason dwCallbackReason,
        IntPtr hSourceFile,
        IntPtr hDestinationFile,
        IntPtr lpData);

        
        enum CopyProgressResult : uint
        {
            PROGRESS_CONTINUE = 0,
            PROGRESS_CANCEL = 1,
            PROGRESS_STOP = 2,
            PROGRESS_QUIET = 3
        }

        enum CopyProgressCallbackReason : uint
        {
            CALLBACK_CHUNK_FINISHED = 0x00000000,
            CALLBACK_STREAM_SWITCH = 0x00000001
        }

        [Flags]
        enum CopyFileFlags : uint
        {
            COPY_FILE_FAIL_IF_EXISTS = 0x00000001,
            COPY_FILE_RESTARTABLE = 0x00000002,
            COPY_FILE_OPEN_SOURCE_FOR_WRITE = 0x00000004,
            COPY_FILE_ALLOW_DECRYPTED_DESTINATION = 0x00000008
        }

        internal void XCopy(string oldFile, string newFile)
        {
            int pbCancel = 0;
            progress = 0;
            CopyFileEx(oldFile, newFile, new CopyProgressRoutine(this.CopyProgressHandler), IntPtr.Zero, ref pbCancel, CopyFileFlags.COPY_FILE_RESTARTABLE);
        }

        long progress = 0;

        internal long Progress
        {
            get { return progress; }
            set { 
                progress = value;
                InvokeOnProgressChanged();
            }
        }

        internal event EventHandler ProgressChanged;

        public delegate void InvokeDelegate();
        
        void OnProgressChanged()
        {
            if (ProgressChanged != null)
                ProgressChanged(this, null);
        }

        void InvokeOnProgressChanged()
        {
            if (theForm.InvokeRequired)
            {
                theForm.Invoke(new InvokeDelegate(OnProgressChanged));
            }
            else
                OnProgressChanged();
        }

        private CopyProgressResult CopyProgressHandler(long total, long transferred, long streamSize, long StreamByteTrans, uint dwStreamNumber, CopyProgressCallbackReason reason, IntPtr hSourceFile, IntPtr hDestinationFile, IntPtr lpData)
        {
            double tran = transferred;
            double tot = total;
            double per = tran / tot * 100.0f;
            long prog = (long)Math.Round(per);

            if (progress < prog)
            {
                progress = prog;
                InvokeOnProgressChanged();
            }
            return CopyProgressResult.PROGRESS_CONTINUE;
        }
    }

    
}
